#!/bin/bash
PARENT=$(ps -o cmd --pid ${PPID} | sed -n 2p)
DAEMON_NAME=$0
OPTS=$@

################################################################################
# Packages
################################################################################
PKG_BASE="ivs-base"
PKG_PYTHON="ivs-python"
PKG_UI="ivs-ui"
PKG_RES="ivs-resource"
PKG_VSWITCH="ivs-vswitch"

################################################################################
# Common Constants
################################################################################
TRUE="True"
FALSE="False"

SRV_STATUS_ACTIVE="active"

OOM_SCORE=-1000

################################################################################
# System Commands
################################################################################
CMD_7Z=/usr/bin/7z
CMD_AWK=/usr/bin/awk
CMD_BUSCTL=/usr/bin/busctl
CMD_CAT=/bin/cat
CMD_CHOWN=/usr/bin/chown
CMD_CP=/bin/cp
CMD_DIRNAME=/usr/bin/dirname
CMD_DPKG=/usr/bin/dpkg
CMD_ECHO=/bin/echo
CMD_EGREP=/usr/bin/egrep
CMD_FIND=/usr/bin/find
CMD_GPG=/usr/bin/gpg
CMD_KILL=/bin/kill
CMD_LN=/bin/ln
CMD_MKDIR=/bin/mkdir
CMD_PGREP=/usr/bin/pgrep
CMD_PKILL=/usr/bin/pkill
CMD_PIDOF=/usr/bin/pidof
CMD_PRINTF=/usr/bin/printf
CMD_PYTHON=/usr/bin/python3
CMD_RSYNC=/usr/bin/rsync
CMD_MODPROBE=/usr/sbin/modprobe
CMD_NGINX=/usr/sbin/nginx
CMD_READLINK=/usr/bin/readlink
CMD_RM=/bin/rm
CMD_SED=/usr/bin/sed
CMD_SLEEP=/bin/sleep
CMD_SORT=/usr/bin/sort
CMD_SYSTEMCTL=/usr/bin/systemctl
CMD_TR=/usr/bin/tr
CMD_XARGS=/usr/bin/xargs

################################################################################
# VS Constants
################################################################################
PKG_NAME="ivs-base"
VS_ROOT="/usr/share/${PKG_NAME}"
VS_ROOT_LINK="/ivs"
PKG_ROOT_LINK="${VS_ROOT_LINK}/IVS-Base"
VS_DATA_DIR="/mnt/IVS"
PKG_PRIVATE_DIR="${VS_DATA_DIR}/.${PKG_NAME}"
VS_WEB="/usr/share/${PKG_UI}/vs"
VS_USR_LIB="${VS_ROOT}/usr/lib"
VS_VAR="${VS_ROOT}/var"
VS_RUN="${VS_VAR}/run"
VS_LOG="${VS_VAR}/log"
VS_TMP="${VS_VAR}/tmp"
VS_LIB="${VS_VAR}/lib"
VS_USR_ETC="${VS_ROOT}/usr/etc"
VS_DAEMON_PATH="${VS_ROOT}/etc/qvsd.d"
VS_DEF="${VS_ROOT}/default"
VS_USER_GROUP="root:root"
VS_PYTHON_PATH="${VS_WEB}"

UI_ROOT="/usr/share/${PKG_UI}"
UI_USR_BIN="${UI_ROOT}/usr/bin"
VS_CONF="${UI_ROOT}/vs.conf"
VS_DB="${UI_ROOT}/vs.db"

BASE_ROOT="/usr/share/${PKG_BASE}"
BASE_USR_BIN="${BASE_ROOT}/usr/bin"
BASE_USR_SBIN="${BASE_ROOT}/usr/sbin"
BASE_RUN="${BASE_ROOT}/var/run"

PYTHON_ROOT="/usr/share/${PKG_PYTHON}"
PYTHON_USR_BIN="${PYTHON_ROOT}/usr/bin"

REDIS_PID="${VS_RUN}/redis.pid"
WORKERS_PID="${VS_RUN}/workers.pid"
UWSGI_PID="${VS_RUN}/uwsgi.pid"

LIBVIRTD_PID="${BASE_RUN}/libvirtd.pid"
VIRTLOGD_PID="${BASE_RUN}/virtlogd.pid"

VS_ISO_DIR="ISOs"
VS_VM_DIR="VMs"
VS_EXPORTED_DIR="Exported"

NGINX_VS_CONF="${VS_ROOT}/usr/etc/nginx.conf"
NGINX_SYS_CONF="/etc/nginx/conf.d/ivec.conf"

UWSGI_CONF="${VS_USR_ETC}/uwsgi.ini"
UWSGI_WARNING_CONF="${VS_USR_ETC}/static.ini"
UWSGI_UNIX_SOCKET="${VS_TMP}/uwsgi.sock"
UWSGI_CONF_SEC_UWSGI="uwsgi"
UWSGI_CONF_SEC_UWSGI_OPT_HTTP="http"
UWSGI_CONF_SEC_UWSGI_OPT_HTTPS="https"
UWSGI_CONF_SEC_SOCKET="socket"

################################################################################
# VS Commands
################################################################################
CMD_QEMU="${BASE_USR_BIN}/qemu-system-x86_64"
CMD_VIRTLOGD="${BASE_USR_SBIN}/virtlogd"
CMD_LIBVIRTD="${BASE_USR_SBIN}/libvirtd"
CMD_VIRSH="${BASE_USR_BIN}/virsh"

CMD_PYTHON="${PYTHON_USR_BIN}/python3"

CMD_UWSGI="${UI_USR_BIN}/uwsgi"

################################################################################
# Log functions
################################################################################
elog() {
    local i
    local def_log_dir=/tmp/qvs
    local log_output

    if [ -e ${VS_LOG} ]; then
        log_output=${VS_LOG}/qvsd.log
    else
        if [ ! -e ${def_log_dir} ]; then
            ${CMD_MKDIR} -p ${def_log_dir}
        fi
        log_output=${def_log_dir}/qvsd.log
    fi

    if [ -f ${log_file} ]; then
        printf "[$(date +"%d/%b/%Y:%H:%M:%S %z")]" >> ${log_output}
    fi

    for i in $@
    do
        printf  " $i" >> ${log_output}
    done
    echo "" >> ${log_output}
}

stderr_log() {
    [ -n "$1" ] && ${CMD_ECHO} "[$1]" >&2
}

qvs_log() {
    run_manage log --message "$1" --data "$2"
}

elog2manage_log() {
    elog "$@" >> ${VS_LOG}/qvs-manage.log 2>&1
}

################################################################################
# D-Bus functions
################################################################################
dbus_call()
{
    local service=$1
    local obj_path=$2
    local inf_path=$3
    local method=$4
    local signatures=$5
    local args=${@:6}

    ${CMD_BUSCTL} call ${service} ${obj_path} ${inf_path} ${method} ${signatures} ${args}
}

################################################################################
# Service functions
################################################################################
get_srv_status()
{
    local srv_name=$1
    local srv_status=$(${CMD_SYSTEMCTL} is-active ${srv_name})
    local ret=$?
    if [ ${ret} -ne 0 ]; then
        ${CMD_ECHO} ${FALES}
        return
    fi
    if [ x"${srv_status}" == x"${SRV_STATUS_ACTIVE}" ]; then
        ${CMD_ECHO} ${TRUE}
        return
    fi
    ${CMD_ECHO} ${FALES}
}

################################################################################
# Utils functions
################################################################################
get_lib_path()
{
    local qvm_pkgs=(${PKG_RES} ${PKG_BASE} ${PKG_PYTHON} ${PKG_UI})
    local len=${#qvm_pkgs[@]}
    local lib_str=''

    for i in ${qvm_pkgs[@]}; do
        pkg_lib=/usr/share/${i}/usr/lib
        if [ x"${i}" == x"${qvm_pkgs[0]}" ]; then
            lib_str=${pkg_lib}
        else
            lib_str=${pkg_lib}:${lib_str}
        fi
    done
    ${CMD_ECHO} ${lib_str}
}

print_env() {
    elog "================="
    elog PKG_NAME=${PKG_NAME}
    elog VS_ROOT=${VS_ROOT}
    elog VS_WEB=${VS_WEB}
    elog "---------"
    elog PATH=${PATH}
    elog LD_LIBRARY_PATH=${LD_LIBRARY_PATH}
    elog PYTHON=${CMD_PYTHON}
    elog PYTHONPATH=${PYTHONPATH}
    elog "---------"
    elog DJANGO_SETTINGS_MODULE=${DJANGO_SETTINGS_MODULE}
    elog "================="
}

setup_env()
{
    local extra=${VS_ROOT}/script/${PKG_NAME}.sh

    if [ -f ${extra} ]; then
        source ${extra}
    fi

    export PATH=${VS_ROOT}/usr/bin:${VS_ROOT}/usr/sbin:${PATH}
    export LD_LIBRARY_PATH=$(get_lib_path):${LD_LIBRARY_PATH}
    export PYTHONPATH=${VS_PYTHON_PATH}
    export DJANGO_SETTINGS_MODULE=vs.settings

    print_env
}

get_redis_pid()
{
    if [ ! -f ${REDIS_PID} ]; then
        ${CMD_ECHO} ""
        return
    fi
    ${CMD_ECHO} $(${CMD_CAT} ${REDIS_PID})
}

get_workers_pid()
{
    local ppid=$(${CMD_CAT} ${WORKERS_PID})
    ${CMD_ECHO} "${ppid} $(${CMD_PGREP} -P ${ppid})"
}

get_uwsgi_pid()
{
    local ppid=$(${CMD_CAT} ${UWSGI_PID})
    ${CMD_ECHO} "${ppid} $(${CMD_PGREP} -P ${ppid})"
}

get_qemu_pid()
{
    ${CMD_ECHO} $(${CMD_PIDOF} -x ${CMD_QEMU})
}

get_libvirt_pid()
{
    ${CMD_ECHO} $(${CMD_CAT} ${LIBVIRTD_PID})
}

get_cmd_by_pid()
{
    local pid=$1
    ${CMD_CAT} /proc/${pid}/cmdline | ${CMD_TR} '\000' ' '
}

___exec_daemon()
{
    . $1 $2 && elog "${1##*-} $2 success code $?" || \
        (stderr_log "${1##*-} $2 failure code $?"; return 1)
}

___exec_group()
{
    local group_name=$1
    local opt=$2
    local opt_start="start"
    local opt_check="check"
    local opt_stop="stop"
    local daemons=$(${CMD_FIND} ${group_name} -maxdepth 1 -type f -o -type l -name "*-*")

    if [ x"${opt}" == x"${opt_start}" ] || [ x"${opt}" == x"${opt_check}" ]; then
        daemons=($(${CMD_ECHO} ${daemons} | ${CMD_XARGS} -n1 | ${CMD_SORT}))
    elif [ x"${opt}" == x"${opt_stop}" ]; then
        daemons=($(${CMD_ECHO} ${daemons} | ${CMD_XARGS} -n1 | ${CMD_SORT} -r))
    fi

    for i in ${!daemons[@]}
    do
        d=${daemons[$i]}
        _d=${d##*/}
        if [ x"${opt}" == x"${opt_start}" ]; then
            elog "\tStarting        ${_d##*-}"
        elif [ x"{opt}" == x"${opt_stop}" ]; then
            elog "\tStopping        ${_d##*-}"
        fi

        ___exec_daemon $d ${opt}

        if [ x"$?" != x"0" ]; then
            if [ x"${opt}" == x"${opt_start}" ]; then
                break
            elif [ x"${opt}" == x"${opt_check}" ]; then
                return 1
            fi
        fi
    done

    if [ $i != $(( ${#daemons[@]} - 1)) ];then
        d=${daemons[$i]}
        _d=${d##*/}
        elog "Group ${group_name##*-} ${opt} failure with ${_d##*-}"

        for (( j=$i; j >= 0; j-- ))
        do
            d=${daemons[$j]}
            elog "Stop ${d##*/}"
            ___exec_daemon $d stop
        done
        return 1
    fi
    return 0
}

__exec_daemon()
{
    if [ -d $1 ]; then
        elog $1 is a group
        ___exec_group $1 $2
    else
        ___exec_daemon $1 $2
    fi
}

__update_oom_score()
{
    local pids=($1)
    for p in ${pids[@]};
    do
        ${CMD_ECHO} ${OOM_SCORE} > /proc/${p}/oom_score_adj
    done
}

update_oom_score()
{
    local redis_pid
    local workers_pid
    local uwsgi_pid

    local libvirtd_pid
    local qemu_pid

    if [ x"${PKG_NAME}" == x"${PKG_UI}" ]; then
        redis_pid=$(get_redis_pid)
        __update_oom_score ${redis_pid}
        if [ -f ${WORKERS_PID} ]; then
            workers_pid=$(get_workers_pid)
            __update_oom_score ${workers_pid}
        fi
        if [ -f ${UWSGI_PID} ]; then
            uwsgi_pid=$(get_uwsgi_pid)
            __update_oom_score ${uwsgi_pid}
        fi
    elif [ x"${PKG_NAME}" == x"${PKG_BASE}" ]; then
        if [ -f ${LIBVIRTD_PID} ]; then
            libvirtd_pid=$(get_libvirt_pid)
            __update_oom_score ${libvirtd_pid}
        fi
        qemu_pid=$(get_qemu_pid)
        __update_oom_score ${qemu_pid}
    fi
}

kill_proc() {
    local signal=$1
    local pids=($2)
    for pid in ${pids[@]}
    do
        elog "Kill PID:[${pid}] ($(ps -p ${pid} -o cmd=)))"
        ${CMD_KILL} ${signal} ${pid}
    done
}

kill_proc_by_pid()
{
    local pid_path=$1
    local signal=$2

    if [ x"${signal}" == x"" ]; then
        signal="-9"
    fi

    if [ ! -f ${pid_path} ]; then
        elog "${pid_path} does not exist"
        return
    fi
    kill_proc "${signal}" $(${CMD_CAT} ${pid_path})
}

create_dir()
{
    local target=$1

    if [ -e ${target} ]; then
        return
    fi
    elog "Create dir: ${target}"
    ${CMD_MKDIR} -p ${target}
}

create_link()
{
    local src=$1
    local dst=$2

    elog "Create link from ${src} to ${dst}"
    ${CMD_LN} -s ${src} ${dst}
}

check_process()
{
    local pid
    local link
    local pid_name=$1
    local proc_name=$2

    if [ ! -f ${VS_RUN}/${pid_name} ]; then
        elog "Process ${proc_name} is not running"
        return 0
    fi

    pid=$(${CMD_CAT} ${VS_RUN}/${pid_name})
    link=$(${CMD_READLINK} /proc/${pid}/exe)
    if [ ! -z "${link}" ] && [ x"${link}" == x"${proc_name}" ]; then
        elog "Process ${proc_name} has alread been started @${pid}"
        return 1
    fi

    elog "Process ${proc_name} is not running"
    remove ${VS_RUN}/${pid_name}
    return 0
}

checkpid_and_rm()
{
    local pid_path=${VS_RUN}/$1
    if [ -f ${pid_path} ]; then
        remove ${pid_path}
    fi
}

remove()
{
    local target=$1
    if [ -d ${target} ]; then
        elog "Remove Dir: ${target}"
    else
        elog "Remove File: ${target}"
    fi
    ${CMD_RM} -rf ${target}
}

################################################################################
# Run commands
################################################################################
__find_manage() {
    ${CMD_ECHO} "-m manage"
}

find_manage=__find_manage

run_command() {
    PATH=$PATH "$@"
}

run_py_command() {
    run_command ${CMD_PYTHON} -u "$@"
}

run_manage_log() {
    elog2manage_log "begin run_manage $@"
    run_py_command $($find_manage) "$@" \
        >> ${VS_LOG}/qvs-manage.log \
        2>&1
    elog2manage_log "end run_manage $@"
}

run_manage() {
    run_manage_log "$@"
}

################################################################################
# Activity functions
################################################################################
pre_start()
{
    elog "Pre Start"
}

start_daemon()
{
    local daemons=$(${CMD_FIND} ${VS_DAEMON_PATH} -maxdepth 1 -name '*-*')
    local _d

    daemons=$(${CMD_ECHO} ${daemons} | ${CMD_XARGS} -n1 | ${CMD_SORT})
    for d in ${daemons}
    do
        _d=${d##*/}
        elog "Checking        ${_d##*-}"
        __exec_daemon $d check
        if [ x"$?" != x"0" ]; then
            elog "Checking failed"
            exit 1
        fi
    done

    for d in ${daemons}
    do
        _d=${d##*/}
        elog "Starting        ${_d##*-}"
        __exec_daemon $d start
        if [ x"$?" != x"0" ]; then
            exit 1
        fi
    done

    update_oom_score
}

stop_daemon()
{
    local daemons=$(${CMD_FIND} ${VS_DAEMON_PATH} -maxdepth 1 -name '*-*')
    local _d

    daemons=$(${CMD_ECHO} ${daemons} | ${CMD_XARGS} -n1 | ${CMD_SORT} -r)

    for d in ${daemons}
    do
        _d=${d##*/}
        elog "Stopping        ${_d##*-}"
        __exec_daemon $d stop
    done
}

post_stop()
{
    elog "Post Stop"
}

restart_web()
{
    __exec_daemon ${VS_DAEMON_PATH}/78-uwsgi stop
    ${CMD_SLEEP} 3
    __exec_daemon ${VS_DAEMON_PATH}/78-uwsgi start
}

list_daemon()
{
    local daemons=$(${CMD_FIND} ${VS_DAEMON_PATH} -maxdepth 1 -name '*-*')
    local _d

    daemons=$(${CMD_ECHO} ${daemons} | ${CMD_XARGS} -n1 | ${CMD_SORT})

    ${CMD_ECHO} "-----------------------------------------------"
    printf "order\tname\t\tsub-order\tsub-name\n"
    for d in ${daemons}
    do
        _d=${d##*/}
        printf "${_d%-*}\t${_d##*-}\n"

        #deamons in group
        if [ -d $d ]; then
            local _daemons=($(${CMD_FIND} $d -maxdepth 1 -type f -o -type l -name '*-*' | sort))
            for __d in ${_daemons[@]}
            do
                _d=${__d##*/}
                printf "\t\t\t%s\t\t%s\n" ${_d%-*} ${_d##*-}
            done
        fi
    done
}

get_status()
{
    ${CMD_ECHO} "-----------------------------------------------"
    ${CMD_ECHO} "QVS status-------------------------------------"
    ${CMD_ECHO} "-----------------------------------------------"
    if [ x"${PKG_NAME}" == x"${PKG_UI}" ]; then
        ${CMD_ECHO} "uwsgi        running @" "$(get_uwsgi_pid)"
        ${CMD_ECHO} "workers      running @" "$(get_workers_pid)"
        ${CMD_ECHO} "redis-server running @" "$(get_redis_pid)"
    elif [ x"${PKG_NAME}" == x"${PKG_BASE}" ]; then
        ${CMD_ECHO} "qemu         running @" "$(get_qemu_pid)"
        ${CMD_ECHO} "libvirtd     running @" "$(get_libvirtd_pid)"
    fi
    ${CMD_ECHO} "-----------------------------------------------"
    ${CMD_ECHO} "${PKG_NAME}    service status:" "$(get_srv_status ${PKG_NAME})"
    ${CMD_ECHO} "-----------------------------------------------"
}

debug()
{
    local dev_name="vsweb-deb"
    ${CMD_DPKG} -s "p7zip-full"
    if [ $? -ne 0 ]; then
        ${CMD_ECHO} "p7zip-full is needed to active debug mode"
        exit 1
    fi
    ${CMD_GPG} --homedir "${VS_ROOT}/.conf/gpg/" --output "${VS_ROOT}/${dev_name}.tar.7z" --decrypt "${VS_ROOT}/${dev_name}"
    ${CMD_7Z} x -so "${VS_ROOT}/${dev_name}.tar.7z" | tar xf - -C "${VS_ROOT}"
    ${VS_ROOT}/qvsd.sh restart_web
    run_manage log_level -S debug
}

################################################################################
# Main function
################################################################################
main()
{
    setup_env
    opt=$1
    case "${opt}" in
        startpre)
            pre_start
            ;;
        start)
            start_daemon
            ;;
        stop)
            stop_daemon
            ;;
        stoppost)
            post_stop
            ;;
        list)
            list_daemon
            ;;
        status)
            get_status
            ;;
        restart)
            $0 stop
            $0 start
            ;;
        restart_web)
            restart_web
            ;;
        debug)
            debug
            ;;
        manage)
            run_py_command $($find_manage) ${@:2}
            ;;
        shell)
            run_command ${@:2}
            ;;
        *)
            ${CMD_ECHO} "Usage: ${DAEMON_NAME} {start|stop|list|status|restart|manage mange_cmd|shell shell_cmd}}"
            exit $1
    esac

    elog "${DAEMON_NAME} ${OPTS} finished (PID=$$, Parent=$(get_cmd_by_pid ${PPID}))"
}

if [ x"${PARENT}" != x"sudo -s" ]; then
    main ${OPTS}
else
    setup_env
fi

