#!/bin/bash

################################################################################
# Commands
################################################################################
CMD_ECHO=/bin/echo
CMD_MKDIR=/bin/mkdir
CMD_SLEEP=/bin/sleep
CMD_SYSTEMCTL=/usr/bin/systemctl

################################################################################
# Constants
################################################################################
FALSE="false"
INTERVAL=10
SRV_UI="ivs-ui"
SRV_BASE="ivs-base"
SRV_STATUS_ACTIVE="active"
TRUE="true"
VS_LOG="/usr/share/ivs-base/var/log"

################################################################################
# Log functions
################################################################################
elog() {
    local i
    local def_log_dir=/tmp/qvs
    local log_output

    if [ -e ${VS_LOG} ]; then
        log_output=${VS_LOG}/start-ui.log
    else
        if [ ! -e ${def_log_dir} ]; then
            ${CMD_MKDIR} -p ${def_log_dir}
        fi
        log_output=${def_log_dir}/start-ui.log
    fi

    if [ -f ${log_file} ]; then
        printf "[$(date +"%d/%b/%Y:%H:%M:%S %z")]" >> ${log_output}
    fi

    for i in $@
    do
        printf  " $i" >> ${log_output}
    done
    echo "" >> ${log_output}
}

################################################################################
# Service functions
################################################################################
get_srv_status()
{
    local srv_name=$1
    local srv_status=$(${CMD_SYSTEMCTL} is-active ${srv_name})
    local ret=$?
    if [ ${ret} -ne 0 ]; then
        ${CMD_ECHO} ${FALSE}
        return
    fi
    if [ x"${srv_status}" == x"${SRV_STATUS_ACTIVE}" ]; then
        ${CMD_ECHO} ${TRUE}
        return
    fi
    ${CMD_ECHO} ${FALSE}
}

start_ui()
{
    local status=$(get_srv_status ${SRV_UI})
    elog "ivs-ui: ${status}"
    if [ x"${status}" == x"${FALSE}" ]; then
        ${CMD_SYSTEMCTL} start ${SRV_UI}
	elog "Start ivs-ui service"
    fi
}


main()
{
    local status=$(get_srv_status ${SRV_BASE})
    while true
    do
        elog "ivs-base: ${status}"
        if [ x"${status}" == x"${TRUE}" ]; then
            start_ui
	    return
        fi
        ${CMD_SLEEP} ${INTERVAL}
	status=$(get_srv_status ${SRV_BASE})
    done
}

main

